package agents;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import utils.Conveyor;

import com.google.soap.search.GoogleSearch;
import com.google.soap.search.GoogleSearchFault;
import com.google.soap.search.GoogleSearchResultElement;

import beans.DataPackage;

/**
 * 
 * This agent makes searches on the Google server, using the Google framework.
 * This agent has two main purposes.
 * GoogleAgent will search for hits on new words that have come in, to keep expanding the growing knowledge-network.
 * GoogleAgent will also verify knowledge that has been found,
 * by making very pointed searches for specific sentences verifying the information.
 * It will then return an estimate on how many hits it found on those pointed searches and thus giving a probability measure of valid information or not.
 * 
 * This agent takes 2 inputs and one output.
 * 
 * The first input is where GoogleAgent receives queries that should be used to find information from Google.
 * The second input is the language-input.  When a new language is set, that information should be sent to the second variable.
 * 
 * The output variable contains the results from Google.
 * 
 * @author Hvard Rykkelid
 *
 */
public class GoogleAgent extends AgentFormat
{
	
	private int noOfPages = 1;
	private GoogleSearch search;
	private static final String agentSettings = "settings" + File.separator +"agentSettings" + File.separator + "GoogleAgent" + File.separator + "settings.txt";
	
	/**
	 * This method will be run when the Thread containing this agentdescription is started
	 */
	public void run()
	{
		// A check to see if the variables have been initialized.
		// If the variables have not been initialized, then this agent will get to it, using a settings-file.
		if(search == null)
		{
			try
			{
				search = new GoogleSearch();
				BufferedReader reader = new BufferedReader(new FileReader(GoogleAgent.agentSettings));
				search.setKey(reader.readLine());
				if(reader.readLine().equals("true")) search.setSafeSearch(true);
				else search.setSafeSearch(false);
				if(reader.readLine().equals("true")) search.setFilter(true);
				else search.setFilter(false);
				noOfPages = Integer.parseInt(reader.readLine());
				reader.close();
			} catch (FileNotFoundException e)
			{
				e.printStackTrace();
			} catch (IOException e)
			{
				e.printStackTrace();
			}
		}
		// This is a loop that every iteration checks the global variable Conveyor.finished if it has been set to true.
		// If it hasn't been set to true, then this agent should still run.
		while(!Conveyor.finished)
		{
			// Checks if there is language information ready that should be acted upon
			if(!input.elementAt(1).isEmpty()) search.setLanguageRestricts(input.elementAt(1).elementAt(0).getData());
			// Checks if there is new information in the inbox.  If there is something it will be processed.
			// If there is no information there, the agent will sleep for a while and make no claim on prosessor-time for a while.
			if(!input.elementAt(0).isEmpty())
			{
				DataPackage local = input.elementAt(0).remove(0);
				// These two next checks are for sending signals about when the whole program should finish.
				// The first one will only pass on the message and continue working.
				if(local.getTag().equals("finish"))
				{
					output.elementAt(0).addElement(local);
					continue;
				}
				// The second one will pass on the message and then terminate this agent.
				if(local.getTag().equals("finished"))
				{
					output.elementAt(0).addElement(local);
					break;
				}
				// Checks if the message-tag equals "verify".
				// If it does it means that GoogleAgent will try to verify some knowledge that has been found
				if(local.getTag().equals("verify"))
				{
					String[] temp = local.getData().split(", ");
					// When verifying some knowledge, GoogleAgent can use many different sentences.
					// The sentence that got the most hits will be the one presented in the final report.
					int max = -1;
					int index = 0;
					for(int i = 5; i < temp.length; i++)
					{
						search.setQueryString('"' + temp[i] + '"');
						// Sometimes there are some problems with searching using google.
						// By using this boolean and the while-sentence with try and catch GoogleAgent will try again until it succeeds if it has trouble searching.
						boolean searched = false;
						while(!searched)
							try
							{
								int tempty = search.doSearch().getEstimatedTotalResultsCount();
								if(max < tempty)
								{
									max = tempty;
									index = i;
								}
								searched = true;
							}
							catch(GoogleSearchFault e)
							{
								System.out.println("Trouble searching!!");
								try
								{
									Thread.sleep(10);
								} catch (InterruptedException e1)
								{
								}
							}
					}
					output.elementAt(0).addElement(new DataPackage(local.getTag() + ", " + temp[4], String.valueOf(max) + ", " + temp[0] + ", " + temp[2] + ", " + temp[3] + ", " + temp[index]));
				}
				// If the tag is not "finish", "finished" or "verify", it means that this is a new word that we want hits for.
				else
				{
					search.setQueryString(local.getData());
					for(int j = 0; j < noOfPages; j++)
					{
						search.setStartResult(10 * j);
						// This is the same technique used above for making sure that the search is made even if there are som initial problems.
						boolean searched = false;
						while(!searched)
							try
							{
								GoogleSearchResultElement[] results = search.doSearch().getResultElements();
								for(int i = 0; i < results.length; i++)
								{
									output.elementAt(0).addElement(new DataPackage(local.getTag().split(", ")[0] + ", " + local.getData() + ", " + results[i].getURL(), results[i].getURL()));
								}
								searched = true;
							} catch (GoogleSearchFault e)
							{
								System.out.println("Trouble searching!!");
								try
								{
									Thread.sleep(10);
								} catch (InterruptedException e1)
								{
								}
							}
					}
				}
			} else try
			{
				Thread.sleep(30);
			} catch (InterruptedException e)
			{
			}
			Thread.yield();
		}
	}	
}