package agents;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Vector;

import javax.swing.JOptionPane;

import utils.Conveyor;

import beans.DataPackage;
import beans.KeySentenceBean;

/**
 * 
 * KnowledgeGatherAgent uses a settings file to identify key-sentences in text and recognize the knowledge there.
 * The settings file can possibly contain settings for multiple languages, but only one language will be kept in memory at once.
 * When KnowledgeGatherAgent identifies knowledge in the text, it will create an output.
 * The output contains information on what kind of knowledge has been identified, what sentences to use to verify it and what the newly discovered word is.
 * 
 * The key-sentences will be ordered so that the longest sentences, in terms of number of words, will be first of the sentences.
 * This makes it possible to search for the longest and hardest-to-match sentences first, and to remove those matches from the text,
 * before the search for the shorter and easier-to-match sentences begins.
 * Often the shorter sentences will be a substring of the longer sentences, and therefor it is important to remove the longer sentences first.
 * 
 * This agent takes 2 inputs and one output.
 * 
 * The first input is where KnowledgeGatherAgent receives text that should be searched for knowledge.
 * The second input is the language-input.  When a new language is set, that information should be sent to the second variable.
 * 
 * The output variable contains the knowledge found, the sentences to use to verify it and the new word to use in search of further knowledge.
 * 
 * @author Hvard Rykkelid
 *
 */
public class KnowledgeGatherAgent extends AgentFormat
{
	private static Vector<KeySentenceBean> sentences;
	
	private static final String agentSettings = "settings" + File.separator +"agentSettings" + File.separator + "KnowledgeGatherAgent" + File.separator + "settings.txt";

	private static String language;

	private static String[] numberIdent;
	
	private static String[][] searchKeys; 

	/**
	 * This method will be run when the Thread containing this agentdescription is started
	 */
	public void run()
	{
		// The reason why this agent is set to sleep initially is because it has nothing to do to begin with,
		// and it has the task of removing the language information from the language inbox.  It shouldn't do this until all have read the information.
		try
		{
			Thread.sleep(100);
		} catch (InterruptedException e)
		{
			e.printStackTrace();
		}
		// Checking if there is some new language information, and if so initializes the static variables based on the language-information
		if(!input.elementAt(1).isEmpty())
		{
			String lang = input.elementAt(1).remove(0).getData();
			if(KnowledgeGatherAgent.language == null || !KnowledgeGatherAgent.language.equals(lang))
			{
				KnowledgeGatherAgent.sentences = new Vector<KeySentenceBean>();
				KnowledgeGatherAgent.language = lang;
				initiateAgent();
			}
		}
		// This is a loop that every iteration checks the global variable Conveyor.finished if it has been set to true.
		// If it hasn't been set to true, then this agent should still run.
		while(!Conveyor.finished)
		{
			// Checks if there is new information in the inbox.  If there is something it will be processed.
			// If there is no information there, the agent will sleep for a while and make no claim on prosessor-time for a while.
			if(!input.elementAt(0).isEmpty())
			{
				// These two next checks are for sending signals about when the whole program should finish.
				// The first one will only pass on the message and continue working.
				if(input.elementAt(0).elementAt(0).getTag().equals("finish"))
				{
					output.elementAt(0).addElement(input.elementAt(0).remove(0));
					continue;
				}
				// The second one will pass on the message and then terminate this agent.
				if(input.elementAt(0).elementAt(0).getTag().equals("finished"))
				{
					output.elementAt(0).addElement(input.elementAt(0).remove(0));
					break;
				}
				DataPackage local = input.elementAt(0).remove(0);
				// Adding a <space> before the text.  This is to make the identifying of words easier, and this <space> will be removed later on.
				String result = " " + local.getData();
				String[] tag = local.getTag().split(", ");
				// This loop checks all the knowledge-identifying senteces and tries to find matches in the text.
				for(int i = 0; i < KnowledgeGatherAgent.sentences.size(); i++)
				{
					int temp = 0;
					int index = 0;
					// This loop finds the first word in the current knowledge-identifying sentece that is not a reference to one of the two key-words
					// It then finds the index of the first occurence of that word in the text received.  This index is stored in "index".
					while(true)
					{
						if(KnowledgeGatherAgent.sentences.elementAt(i).getWords()[temp].charAt(0) != '-')
						{
							index = result.indexOf(KnowledgeGatherAgent.sentences.elementAt(i).getWords()[temp]);
							while(index > 1 && temp > 0)
							{
								index = result.lastIndexOf(" ", index - 2) + 1;
								temp--;
							}
							break;
						}
						temp++;
					}
					// This loop will use the current knowledge-identifying sentece and try and match it up with the content of the received text.
					// If the index returned after a run equals -1, then controll will be returned and the knowledge-identifying sentence will be changed.
					while(index > -1)
					{
						// "words" is an array of all the words after the index of "index" from "result".
						String[] words = result.substring(index).split(" ");
						int elementNumber = 0;
						String a = "";
						String b = "";
						int number = 0;
						String arity = "";
						// Here it is checked that the counter "number" does not exceed the number of words in "words".
						// "number" is an index for "words", while "elementNumber" is an index for the current key-sentence.
						while(number < words.length)
						{
							// Checks if the current word from "words" should be matched up as one of the key-words from the key-sentence.
							if(KnowledgeGatherAgent.sentences.elementAt(i).getWords()[elementNumber].equals("-a-"))
							{
								if(words[number].length() < 3)
									if(number == 0) break;
									else number++;
								else
								{
									a = words[number];
									elementNumber++;
									number++;
								}
							}
							else if(KnowledgeGatherAgent.sentences.elementAt(i).getWords()[elementNumber].equals("-b-"))
							{
								if(words[number].length() < 3)
									if(number == 0) break;
									else number++;
								else
								{
									b = words[number];
									elementNumber++;
									number++;
								}
							}
							// Checks if an element in "words" should be a number to match up with the description of the key-sentence.
							else if(KnowledgeGatherAgent.sentences.elementAt(i).getWords()[elementNumber].equals("#"))
							{
								String tempString = "";
								while(isaNumber(words[number]))
								{
									tempString += words[number];
									number++;
								}
								if(!tempString.equals(""))
								{
									//TODO identify numbers
								}
								else break;
								elementNumber++;
							}
							else
							{
								// If the word from "words" does not match the word found in the key-sentence, the loop is aborted.
								// This forces the agent to find a new index and start over matching the current key-sentence with the text at this new index.
								if(!words[number].equals(KnowledgeGatherAgent.sentences.elementAt(i).getWords()[elementNumber])) break;
								// If the agent has come to this part of the code, it means that so far the text is matching up with the key-sentence.
								// That means it is time to match the next element.
								elementNumber++;
								number++;
							}
							// If this check is true, then it means that we have found a match for the key-sentence in the text.
							if(elementNumber == KnowledgeGatherAgent.sentences.elementAt(i).getWords().length)
							{
								// This checks that the elements stored in "a" and "b" are not the same.
								if(!a.equals(b))
								{
									// Building the "tempSentence" which will be used to verify the truth of the knowledge found.
									String tempSentence = "";
									for(int j = 0; j < KnowledgeGatherAgent.searchKeys[0].length; j++)
										if(KnowledgeGatherAgent.searchKeys[0][j].equals(KnowledgeGatherAgent.sentences.elementAt(i).getLabel()))
										{
											tempSentence = KnowledgeGatherAgent.searchKeys[1][j].replaceAll("-a-", a).replaceAll("-b-", b);
											break;
										}
									// Checks whether the word the system is currently searching for has been stored in either "a" or "b".
									// If the word the system is searching for is not in one of those two variables, then no interesting information has been found.
									// If the word is found in one of those two variables, the agent will create an output with the knowledge found, a sentence for verification and a new interesting word to research.
									if(a.equals(tag[1]))
									{
										if(arity.equals("")) output.elementAt(0).addElement(new DataPackage((Integer.parseInt(tag[0]) - 1) + ", " + tag[1] + ", " + tag[2] + ", " + KnowledgeGatherAgent.sentences.elementAt(i).getLabel() + "(" + a + "," + b + ")" + ", " + tempSentence, b));
										else output.elementAt(0).addElement(new DataPackage((Integer.parseInt(tag[0]) - 1) + ", " + tag[1] + ", " + tag[2] + ", " + KnowledgeGatherAgent.sentences.elementAt(i).getLabel() + "(" + a + "," + b + "," + arity + ")" + ", " + tempSentence, b));
									}
									else if(b.equals(tag[1]))
									{
										if(arity.equals("")) output.elementAt(0).addElement(new DataPackage((Integer.parseInt(tag[0]) - 1) + ", " + tag[1] + ", " + tag[2] + ", " + KnowledgeGatherAgent.sentences.elementAt(i).getLabel() + "(" + a + "," + b + ")" + ", " + tempSentence, a));
										else output.elementAt(0).addElement(new DataPackage((Integer.parseInt(tag[0]) - 1) + ", " + tag[1] + ", " + tag[2] + ", " + KnowledgeGatherAgent.sentences.elementAt(i).getLabel() + "(" + a + "," + b + "," + arity + ")" + ", " + tempSentence, a));
									}
								}
								break;
							}
						}
						// This loop will try and find an index for the next round of matching sentences.  If none is found "index" will be returned with -1 as value.
						temp = 0;
						while(true)
						{
							// Finding the first normal word in the key-sentence
							if(KnowledgeGatherAgent.sentences.elementAt(i).getWords()[temp].charAt(0) != '-')
							{
								// If "elementNumber" points to the last element of the present key-sentence, that means that it has found some knowledge and a match.
								// If that is the case, then this sentence should be deleted from the text,
								// so that the shorter key-sentences, which may be substrings of the longer key-sentences, will not identify the wrong words as key-words.
								if(elementNumber == KnowledgeGatherAgent.sentences.elementAt(i).getWords().length)
								{
									// These two try-sentences tries to build a text that does not contain the found knowledge-sentence.
									String tempty = "";
									try
									{
										tempty = result.substring(0, index);
									}
									catch(Exception e)
									{
									}
									try
									{
										tempty += result.substring(result.indexOf(words[number], index) + words[number].length() + 1);
									}
									catch(Exception e)
									{
									}
									// Clearing up extra <space> which may have been produced by trying to cut out the knowledge-giving sentence.
									result = tempty.replaceAll("  ", " ");
								}
								else
								{
									// If there hasn't been identified a knowledge-giving sentence,
									// then "index" will be set to point right behind where it last found a match for the first word in the key-sentence.
									int j = 0;
									while(temp >= j && index > 0)
									{
										index = result.indexOf(" ", index) + 1;
										j++;
									}
								}
								// Sets "index" to point to first occurence of the first word in the key-sentence after the place where "index" is poining now.
								if(index > 0) index = result.indexOf(KnowledgeGatherAgent.sentences.elementAt(i).getWords()[temp], index);
								// Traces "index" back to where the first element in the current key-sentence should be found,
								// so that "index" points to "-a-" or "-b-" in the key-sentence if that is the first element.
								// This is however made in such a way, so that it will find the index for the first element no matter the order of elements.
								while(temp > 0)
								{
									if(index < 2) break;
									index = result.lastIndexOf(" ", index - 2) + 1;
									temp--;
								}
								// Sometimes the index will end up with having zero as value, but this is not a real or valid value.
								// Therefor this statement will clean this up by making sure that if "index" has value below 1 it will be set to -1.
								if(index < 1) index = -1;
								break;
							}
							temp++;
						}
					}
				}
			} else try
			{
				Thread.sleep(30);
			} catch (InterruptedException e)
			{
			}
			Thread.yield();
		}
	}
	
	/**
	 * 
	 * This method checks whether the input-string symbolizes a number or not.
	 * The input-string can contain numbers in numbers-format or text-format.
	 * Recognition of numbers in text-format relies on language-specific information from the settings-file.
	 * 
	 * @param string - the String to check
	 * @return true if the String was a number
	 */
	private boolean isaNumber(String string)
	{
		for(int i = 0; i < KnowledgeGatherAgent.numberIdent.length; i++)
			if(string.startsWith(KnowledgeGatherAgent.numberIdent[i])) return true;
		return false;
	}

	/**
	 * This method initiates the variables of KnowledgeGatherAgent.  It reads information from the settingsfile.
	 * All the information is language-specific, and the agent will only fetch information relevant to the current language.
	 * The elements of information fetched are: How to identify numbers in this language, how to verify knowledge in this language
	 * and how to identify knowledge in this language.
	 *
	 */
	private void initiateAgent()
	{
		try
		{
			BufferedReader reader = new BufferedReader(new FileReader(KnowledgeGatherAgent.agentSettings));
			String content = "";
			while(reader.ready()) content += reader.readLine() + "\n";
			reader.close();
			content = content.toLowerCase().replaceAll("\t", "");
			// This picks out the content for the current language.
			content = content.substring(content.indexOf("<" + KnowledgeGatherAgent.language + ">") + 10, content.indexOf("</" + KnowledgeGatherAgent.language + ">"));
			// This picks out and stores the information on how to identify numbers.
			KnowledgeGatherAgent.numberIdent = content.substring(content.indexOf("<numbers>") + 10, content.indexOf("</numbers>")).split("\n");
			// This removes the information of how to store numbers from "content".
			content = content.substring(0, content.indexOf("<numbers>")) + content.substring(content.indexOf("</numbers>") + 10, content.length() - 1);
			// This picks out the information on how to verify sentences and stores the information in "temp".
			String[] temp = content.substring(content.indexOf("<keys>") + 7, content.indexOf("</keys>")).split("\n");
			// Initializing the variable "KnowledgeGatherAgent.searchKeys" with a size corresponding to the amount of information about verifying knowledge.
			KnowledgeGatherAgent.searchKeys = new String[2][temp.length];
			// Stores information on how to verify knowledge in "KnowledgeGatherAgent.searchKeys"
			for(int i = 0; i < temp.length; i++)
			{
				String[] tempty = temp[i].split(" = ");
				if(tempty.length > 1)
				{
					KnowledgeGatherAgent.searchKeys[0][i] = tempty[0];
					KnowledgeGatherAgent.searchKeys[1][i] = tempty[1];
				}
			}
			// Removing the content about how to verify knowledge from "content".
			// This should make sure that all the information that is left is all about identifying knowledge in text.
			content = content.substring(0, content.indexOf("<keys>")) + content.substring(content.indexOf("</keys>") + 7, content.length() - 1);
			// As long as there is information left in "content", this loop will try to extract and store that information.
			while(content.length() > 3)
			{
				// "label" will contain the name of the first tag found in "content".
				String label = content.substring(content.indexOf("<") + 1, content.indexOf(">"));
				// "temp" will contain the information between the start-tag and the end-tag of "label".
				temp = content.substring(content.indexOf("<" + label + ">") + 3 + label.length(), content.indexOf("</" + label + ">")).split("\n");
				// This for-sentence will store a key-sentence found in "temp", with the sentences with the most words stored first.
				for(int i = 0; i < temp.length; i++)
				{
					int j = 0;
					boolean placed = false;
					while(j < KnowledgeGatherAgent.sentences.size() && !placed)
					{
						if(KnowledgeGatherAgent.sentences.elementAt(j).getWords().length < temp[i].length())
						{
							KnowledgeGatherAgent.sentences.insertElementAt(new KeySentenceBean(label, temp[i]), j);
							placed = true;
						}
						j++;
					}
					if(!placed) KnowledgeGatherAgent.sentences.addElement(new KeySentenceBean(label, temp[i]));
				}
				content = content.substring(content.indexOf("</" + label + ">") + 3 + label.length());
			}
		} catch (FileNotFoundException e)
		{
			JOptionPane.showMessageDialog(null, "Settings-file for KnowledgeGatherAgent is not found. File-location should be " + (new File(KnowledgeGatherAgent.agentSettings)).getAbsolutePath());
			System.exit(0);
		} catch (IOException e)
		{
			JOptionPane.showMessageDialog(null, "Trouble reading from settings-file for KnowledgeGatherAgent. File-location should be " + (new File(KnowledgeGatherAgent.agentSettings)).getAbsolutePath());
			System.exit(0);
		} catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, "Trouble initiating KnowledgeGatherAgent. Check settings-file at " + (new File(KnowledgeGatherAgent.agentSettings)).getAbsolutePath());
			System.exit(0);
		}
	}
}