package agents;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Vector;

import javax.swing.JOptionPane;

import utils.Conveyor;

import beans.DataPackage;

/**
 * 
 * SearchApprovalAgent keeps track of which searches are conducted and how they go.
 * SearchApprovalAgent receives all search-requests and decides which the system will follow up on,
 * based on previous searches and their results.
 * This means that SearchApprovalAgent receives alot of trafic and that it has to analyze this trafic,
 * and direct dataflow to some extent.
 * 
 * This agent takes 2 inputs and at least 4 outputs.
 * 
 * The first input is the input from the user.  This input tells the agent what word to start a search for.
 * But this first input is also used by other agents to give information on what words to start a search for.
 * Input from these agents may also contain knowledge and knowledge-giving statements to verify.
 * 
 * The second input is the input from the agents working with search-agents, like Google.
 * This input may receive either suggestions on URLs or results from verification of knowledge.
 * 
 * The first output is given the accepted URLs for download of content.
 * The second output is the found knowledge, ready for parsing and presentation (but not yet cleansed of low probability data).
 * The third output is where SearchApprovalAgent can send messages to the user.
 * The number of outputs after three are completely dependent on the number of different agents used for interfacing with search-engines.
 * All the outputs after three will be given the same data and treated as agents interfacing with search-engines.
 * 
 * @author Hvard Rykkelid
 *
 */
public class SearchApprovalAgent extends AgentFormat
{

	// These two vectors store information on which words have been used in a search,
	// and which knowledge-giving statements have been sent to verification.
	private static Vector<String> investigatedWords, verified;
	
	// Keeps track of what searches have been conducted and what the host of that URL was.
	// This agent will keep track of which hosts have given knowledge-giving statements,
	// and will start ignoring hosts that appear alot and does not often give interesting information.
	private static Vector<DataPackage> searches;
	
	// This variable keeps track of the total number of searches being conducted, to help figure out which hosts should be ignored.
	private static int searchCounter;
	
	// This variable keeps track of the total number of knowledge-giving sentences found, to help figure out which hosts should be ignored.
	private static int hitCounter;

	// This variable makes sure that all the instances of SearchApprovalAgent will stop sending request for more searches at the appropriate time.
	private static boolean finished;
	
	/**
	 * This method will be run when the Thread containing this agentdescription is started
	 */
	public void run()
	{
		// Initializing variables
		if(SearchApprovalAgent.investigatedWords == null)
		{
			SearchApprovalAgent.investigatedWords = new Vector<String>();
			SearchApprovalAgent.verified = new Vector<String>();
			SearchApprovalAgent.searches  = new Vector<DataPackage>();
			SearchApprovalAgent.searchCounter = 0;
			SearchApprovalAgent.hitCounter = 0;
			SearchApprovalAgent.finished = false;
		}
		// This is a loop that every iteration checks the global variable Conveyor.finished if it has been set to true.
		// If it hasn't been set to true, then this agent should still run.
		while(!Conveyor.finished)
		{
			// Since this agent has two inputs, it means that the agent should only sleep if none of them have new data.
			// This variable will be set to true if one of them has had any new data during this run.
			boolean activity = false;
			// Checks if there is new information in the first inbox.  If there is something it will be processed.
			if(!input.elementAt(0).isEmpty())
			{
				activity = true;
				DataPackage local = input.elementAt(0).remove(0);
				// Sorting the information received into managable pieces.
				String[] tag = local.getTag().split(", ");
				// These two next checks are for sending signals about when the whole program should finish.
				// The first will modify the message from "finish" to "finished" and send the new message on and continue working.
				if(local.getTag().equals("finish"))
				{
					local.setTag("finished");
					for(int i = 3; i < output.size(); i++) output.elementAt(i).addElement(local);
					continue;
				}
				// The second one will signal to the agent that it is about time to clean up the results and finish the program.
				if(local.getTag().equals("finished"))
				{
					removeUncertainOutputs();
					break;
				}
				// If the number of elements in the tag exceeds 2, then it is a signal that this input is not an initial request.
				// This means that the tag will probably also contain other data which should be acted upon, not just a request for a new search.
				if(tag.length > 2)
				{
					SearchApprovalAgent.hitCounter++;
					// Extracts the host-name from the URL for matching with stored host-names.
					String temp = "";
					try
					{
						temp = new URL(tag[2]).getHost();
					} catch (MalformedURLException e)
					{
					}
					// Matching the host-name of the URL with stored host-names, and so creating data on which URLs give good results and which give bad results.
					// Every result from a host-name rewards the host with 5 points.
					for(int i = 0; i < SearchApprovalAgent.searches.size(); i++)
					{
						String[] tempTag = SearchApprovalAgent.searches.elementAt(i).getTag().split(", ");
						if(tempTag[0].equals(temp))
						{
							SearchApprovalAgent.searches.elementAt(0).setData(String.valueOf(Integer.parseInt(SearchApprovalAgent.searches.elementAt(0).getData()) + 5));
							break;
						}
					}
					// Checking if the data found have allready been verified.
					boolean found = false;
					for(int i = 0; i < SearchApprovalAgent.verified.size(); i++)
						if(SearchApprovalAgent.verified.elementAt(i).equals(local.getData()))
						{
							found = true;
							break;
						}
					// If the data hasn't been verified, start verification.
					if(!found)
					{
						SearchApprovalAgent.verified.addElement(local.getData());
						DataPackage tempty = new DataPackage("verify", local.getData() + ", " + local.getTag());
						for(int i = 3; i < output.size(); i++) output.elementAt(i).addElement(tempty);
					}
				}
				// Checks whether SearchApprovalAgent has allready decided to stop sending out new words for searching.
				if(!SearchApprovalAgent.finished)
				{
					// If the generation-number on the tag is less than 1 it means that SearchApprovalAgent should stop sending new words for searching.
					// It also means that the system is getting close to finishing alltogether.
					// To signal when the system should shut-down, SearchApprovalAgent will send two messages through the whole system loop.
					// The first message ("finish") is just sent on by all agents, but when it comes back it means that all inboxes have been visited.
					// This means that the inboxes are emptying, if not allready empty.
					// To get as much as possible of the information still left, SearchApprovalAgent modifies the tag from "finish" to "finished".
					// Then this new tag goes around the loop once more and when it is received by SearchApprovalAgent again, it is taken as a signal to mop up and finish.
					if(Integer.parseInt(tag[0]) < 1)
					{
						SearchApprovalAgent.finished = true;
						for(int i = 3; i < output.size(); i++) output.elementAt(i).addElement(new DataPackage("finish", ""));
					}
					else
					{
						// Checks whether the word has allready been investigated.
						boolean found = false;
						for(int i = 0; i < SearchApprovalAgent.investigatedWords.size(); i++)
							if(SearchApprovalAgent.investigatedWords.elementAt(i).equals(local.getData()))
							{
								found = true;
								break;
							}
						// If the word has not been investigated, SearchApprovalAgent will call for a new search with that word.
						if(!found)
						{
							SearchApprovalAgent.investigatedWords.addElement(local.getData());
							for(int i = 3; i < output.size(); i++) output.elementAt(i).addElement(new DataPackage(tag[0], local.getData()));
						}
					}
				}
			}
			// Checks if there is new information in the second inbox.  If there is something it will be processed.
			if(!input.elementAt(1).isEmpty())
			{
				activity = true;
				DataPackage local = input.elementAt(1).remove(0);
				// Sorting the information received into managable pieces.
				String[] tag = local.getTag().split(", ");
				// These two next checks are for sending signals about when the whole program should finish.
				// Both messages will only be passed on and not acted upon.  The acting happens elsewhere from the other input.
				if(local.getTag().equals("finish")) output.elementAt(0).addElement(local);
				else if(local.getTag().equals("finished")) output.elementAt(0).addElement(local);
				// Checks whether the tag is "verify", in which the data is results from verifying some knowledge.
				else if(tag[0].equals("verify"))
				{
					// Some printout with information on the data that has been verified.
					System.out.println("verify:");
					System.out.println("tag:");
					System.out.println(local.getTag());
					System.out.println("data:");
					System.out.println(local.getData());
					System.out.println();
					// If the output to the user-interface has been initialized, there will be posted a message there.
					if(output.elementAt(2).size() > 0)
					{
						String temp[] = local.getData().split(", ");
						output.elementAt(2).elementAt(0).setData("Verifying: " + temp[4]);
					}
					// If the verified knowledge is not allready stored, this will store the verified knowledge in the second output.
					boolean exists = false;
					for(int i = 0; i < output.elementAt(1).size(); i++)
						if(output.elementAt(1).elementAt(i).getTag().equals(tag[1]))
						{
							exists = true;
							break;
						}
					if(!exists) output.elementAt(1).addElement(new DataPackage(tag[1], local.getData()));
				}
				else
				{
					// If the tag is not "verify", then the DataPackage is supposed to contain URL-results.
					// This code picks out the host from the URL.
					String temp = "";
					try
					{
						temp = new URL(tag[2]).getHost();
					} catch (MalformedURLException e)
					{
						e.printStackTrace();
					}
					// Tries to find a match for the host-name in previously stored searches.
					boolean found = false;
					for(int i = 0; i < SearchApprovalAgent.searches.size(); i++)
					{
						String[] tempTag = SearchApprovalAgent.searches.elementAt(i).getTag().split(", ");
						if(tempTag[0].equals(temp))
						{
							// When a match is found, this part uses an algorithm to figure out if SearchApprovalAgent should approve a search of that host.
							// The algorithm is that you take the points of the host-name (which really is 5 times the number of hits found using that host).
							// You divide these points on the number of searches that have been made with that host-name.
							// Then you take this number and add the total number of searches made and subtract the total number of results returned.
							// If the result is a number bigger than -10 then the URL is accepted and sent on.
							// This algorithm have not been properly studied with lots of searches to see how it works, but there is some logic behind it.
							// Adding the total number of searches to the number gives you a sort of a buffer, which gives the program time to find results.
							// When there are many searches it is probable that it will take some time to find results for each search.
							// Subtracting the total amount of results makes sure that you adjust the criteria for acceptance of URLs.
							// If there are few results you want to include every result you get.  If you have alot of results you don't need to investigate all sources.
							// Taking the number of hits (times five) on this host-name and dividing it on how many searches have been made on this host-name
							// gives the deciding push as to who should be accepted and who should not be accepted if there are many results and
							// not all sources are needed to produce results.
							found = true;
							if((Integer.parseInt(SearchApprovalAgent.searches.elementAt(i).getData()) / Integer.parseInt(tempTag[1])) + SearchApprovalAgent.searchCounter - SearchApprovalAgent.hitCounter > -10)
							{
								SearchApprovalAgent.searches.elementAt(i).setTag(tempTag[0] + ", " + (Integer.parseInt(tempTag[1]) + 1));
								output.elementAt(0).addElement(local);
								SearchApprovalAgent.searchCounter++;
							}
							// When URLs are rejected, SearchApprovalAgent prints a message
							else System.out.println("URL rejected: " + SearchApprovalAgent.searches.elementAt(i).getTag());
							break;
						}
					}
					// Stores the host-name for future reference if not previously stored.
					if(!found)
					{
						SearchApprovalAgent.searches.addElement(new DataPackage(temp + ", 1", "0"));
						output.elementAt(0).addElement(local);
						SearchApprovalAgent.searchCounter++;
					}
				}
			}
			if(!activity) try
			{
				Thread.sleep(20);
			} catch (InterruptedException e)
			{
			}
			Thread.yield();
		}
	}

	/**
	 * 
	 * This method is run right before the system ends.
	 * This method clears away the knowledge which has had poor results after verification.
	 *
	 */
	private void removeUncertainOutputs()
	{
		if(output.elementAt(1).size() > 0)
		{
			// First this method counts the total number of verified hits on all the verified knowledge.
			int hitCounter = 0;
			for(int i = 0; i < output.elementAt(1).size(); i++)
				if(!output.elementAt(1).elementAt(i).equals("finished"))
					if(Integer.parseInt(output.elementAt(1).elementAt(i).getData().split(", ")[0]) < 1000) hitCounter += Integer.parseInt(output.elementAt(1).elementAt(i).getData().split(", ")[0]);
					else hitCounter += 1000;
			// Then the average hit per result is calculated.
			int avgHitPerResult = hitCounter / output.elementAt(1).size();
			// Then all the verified knowledge is checked if they received more verifying hits than 3% of the average number of hits.
			// This is a pretty simple algorithm, and it would probably be a good idea to make a more advanced one.
			int j = 0;
			while(j < output.elementAt(1).size())
			{
				if(Integer.parseInt(output.elementAt(1).elementAt(j).getData().split(", ")[0]) < (avgHitPerResult * 0.03)) output.elementAt(1).remove(j);
				else j++;
			}
			// Here the global variable "Conveyor.finished" is set to true.  This tells all agents that the program is finished with a run.
			Conveyor.finished = true;
			// The result of the search is shown to the user with a message giving the verification-sentences.
			String message = "";
			for(int i = 0; i < output.elementAt(1).size(); i++)
				message += output.elementAt(1).elementAt(i).getData().split(", ")[4] + "\n";
			JOptionPane.showMessageDialog(null, message);
		}
	}
}