package utils;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Vector;

import javax.swing.JOptionPane;

import agents.AgentFormat;

import beans.DataPackage;

/**
 * 
 * The "Conveyor"-class is a class made to start up all the agents.
 * Conveyor is made in a very general way to make it easier to use this class in connection with other MAS projects.
 * Conveyor reads all its instructions from a settings-file.
 * This settings-file decides which agents should be started and how many should be started of each type, and their priority.
 * In addition the settings-file contains a list of names for input-variables and output-variables for each agent.
 * If a variable-name has not been used before, that variable is created and sent to the agent.
 * When a variable-name has been used before, a reference to that same variable is sent to the agent.
 * That means that agents called by Conveyor does not have to know of each other's existance.
 * All they know is that they receive certain input and output variables and they use them as they are supposed to.
 * Where the messages go and what is done with them is totally up to Conveyor and the rest of the agents.
 * Conveyor has the following important variables:
 * 
 * "data" - This is a variable that contains all the variables that are used by the agents to send and receive information between themselves.
 * Since the variables that the agents use are Vectors with DataPackages, then since it is not known how many such variables are needed,
 * the resulting structure of "data" is Vector<Vector<DataPackage>>.
 * This makes sure that it is possible with an unknown number of variables with an unknown number of elements in each.
 * 
 * "index" - This is variable that contains the names of all the variables in "data", so that a name can be linked to the corresponding variable.
 * 
 * "agents" - Is a Vector of Threads.  This variable contains all the Threads of the agents.
 * 
 * "finished" - Is a boolean variable, and it is a global static variable that is intended to act as a signal to all agents when they are supposed to finish.
 * All agents are recommended to check this variable now and then to figure out when to stop.
 * 
 * "agentSettings" - Is a String that stores the location of the settingsfile for Conveyor.
 * The location for the settingsfile is sent in with the constructor for Conveyor.
 * This means that you can refer to different settings-files or that you can make some part of the program write the file before sending a reference.
 * 
 * 
 * @author Hvard Rykkelid
 *
 */
public class Conveyor
{
	
	private static Vector<Vector<DataPackage>> data;
	
	public static boolean finished = false;
	
	private static Vector<String> index;

	private Vector<Thread> agents;
	
	private String agentSettings = "";
	
	/**
	 * 
	 * The constructor
	 * 
	 * @param agentSettings - the filename for the settings that are to be used
	 */
	public Conveyor(String agentSettings)
	{
		this.agentSettings = agentSettings;
	}
	
	/**
	 * 
	 * A method to change the stored location of the settings-file
	 * 
	 * @param agentSettings - the filename for the settings that are to be used
	 */
	public void setAgentSettings(String agentSettings)
	{
		this.agentSettings = agentSettings;
	}
	
	/**
	 * 
	 * A method to reset the system for a new search.
	 * This method refers to the previously stored name of the settings-file.
	 *
	 */
	public void resetAllData()
	{
		// Initializes/cleans the current variables.
		Conveyor.data = new Vector<Vector<DataPackage>>();
		Conveyor.index = new Vector<String>();
		agents = new Vector<Thread>();
		// Tries to read the information in the settings-file
		try
		{
			BufferedReader reader = new BufferedReader(new FileReader(agentSettings));
			String content = "";
			while(reader.ready()) content += reader.readLine() + "\n";
			reader.close();
			// Cleans up the content
			content = content.replaceAll("\t", "").replaceAll(" ", "");
			// Picks out the agent-specifications from the xml-file, and splits the information into managable pieces.
			String agentSpecs[] = content.substring(content.indexOf("<agents>") + 9, content.indexOf("</agents>")).split("\n");
			// This loop goes through all the agent-specifications and acts upon them
			for(int j = 0; j < agentSpecs.length; j++)
			{
				// Splits the lines of information into more manageable pieces
				String[] categories = agentSpecs[j].split(";");
				String[] tempInputs = categories[3].split(",");
				String[] tempOutputs = categories[4].split(",");
				// Creates an input Vector and an output Vector that will contain all the input-variables and output-variables of the agent.
				Vector<Vector<DataPackage>> inputs = new Vector<Vector<DataPackage>>();
				Vector<Vector<DataPackage>> outputs = new Vector<Vector<DataPackage>>();
				// Adds the proper variables to the input and output variables for the current agent.
				for(int i = 0; i < tempInputs.length; i++) inputs.addElement(getVectorWithName(tempInputs[i]));
				for(int i = 0; i < tempOutputs.length; i++) outputs.addElement(getVectorWithName(tempOutputs[i]));
				// This loop makes sure that Conveyor starts up as many of this type of agent as is set in the settings-file.
				for(int i = 0; i < Integer.parseInt(categories[1]); i++)
				{
					// Gets the corresponding class for the name of the agent in the settings-file
					AgentFormat temp = (AgentFormat) Class.forName(categories[0]).newInstance();
					// Sets the variables of the agent
					temp.init(inputs, outputs);
					// Creates the new Thread with the agent-description
					Thread tempty = new Thread(temp);
					// Sets Thread settings
					tempty.setDaemon(true);
					tempty.setPriority(Integer.parseInt(categories[2]));
					// Adds the new Thread to "agents"
					agents.addElement(tempty);
				}
			}
		} catch (FileNotFoundException e)
		{
			JOptionPane.showMessageDialog(null, "No settingsfile found");
			e.printStackTrace();
			System.exit(0);
		} catch (IOException e)
		{
			JOptionPane.showMessageDialog(null, "IOException");
			e.printStackTrace();
		} catch (InstantiationException e)
		{
			JOptionPane.showMessageDialog(null, "InstantiationException");
			e.printStackTrace();
		} catch (IllegalAccessException e)
		{
			JOptionPane.showMessageDialog(null, "IllegalAccessException");
			e.printStackTrace();
		} catch (ClassNotFoundException e)
		{
			JOptionPane.showMessageDialog(null, "One or more of the agent-definitions could not be resolved");
			e.printStackTrace();
			System.exit(0);
		} catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage());
			System.exit(0);
		}
	}
	
	/**
	 * 
	 * This is a method that should only be called after there has been made an instance of Conveyor.
	 * This method will receive a variable-name.  It will then check if that name has allready been decleared.
	 * If the name has allready been declared, then this method will return the corresponding variable.
	 * If the name has not been declared, it will now be declared and a new variable will be created, before that variable is returned.
	 * This makes sure that after there has been made an instance of Conveyor you can always send in a name to this method.
	 * This method will then always return a variable, either the corresponding or a new one which then will have that name.
	 * 
	 * @param name - the name of the variable you want returned
	 * @return a Vector<DataPackage> variable with the same name as you sent in
	 */
	public static Vector<DataPackage> getVectorWithName(String name)
	{
		if(Conveyor.index.contains(name)) return Conveyor.data.elementAt(Conveyor.index.indexOf(name));
		Vector<DataPackage> temp = new Vector<DataPackage>();
		Conveyor.data.addElement(temp);
		Conveyor.index.addElement(name);
		return temp;
	}
	
	/**
	 * 
	 * This method signals that all the agents should be started and set to work.
	 *
	 */
	public void start()
	{
		Conveyor.finished = false;
		for(int i = 0; i < agents.size(); i++)
		{
			if(!agents.elementAt(i).isAlive()) agents.elementAt(i).start();
		}
	}
}