package utils;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.Vector;

/**
 * 
 * The TreeNode-class is made to simplify storing and searching for words that should be filtered away.
 * Each instance of TreeNode contains its own methods for recursively storing
 * and recursively checking if a word matches the description of a word that should be removed.
 * Each node contains only one letter, but it also contains links to a number of new nodes.
 * If one of those nodes contains the string "end" instead of a letter, it means that if a word that
 * is being checked ends with the letter at the present node, then it is a word that should be removed.
 * Each node has links to other nodes and those nodes will contain other valid letters
 * to follow the allready established trail of letters.
 * If the word that is being checked has a letter following the current node that doesn't exist, 
 * then the word is not to be removed, it is not an unwanted word.
 * 
 * @author Hvard Rykkelid
 *
 */
public class TreeNode
{
	private static final String nodeSettings = "settings" + File.separator +"agentSettings" + File.separator + "FilterAgent" + File.separator + "nodeSettings.txt";
	private static Vector<String> wordEnds;
	private Vector<TreeNode> nodes;
	private String data;
	
	/**
	 * 
	 * The constructor stores the first letter of the String received and sends the rest of the word to be stored in following nodes.
	 * If the String is empty, then the node will store "end".
	 * 
	 * @param data - the received word or part-of-word
	 */
	public TreeNode(String data)
	{
		// Checks whether the variable "TreeNode.wordEnds" have been initialized
		// "TreeNode.wordEnds" is a Vector of Strings containing description of symbols that signify the end of a word
		if(TreeNode.wordEnds == null)
		{
			TreeNode.wordEnds = new Vector<String>();
			try
			{
				BufferedReader reader = new BufferedReader(new FileReader(TreeNode.nodeSettings));
				while(reader.ready()) TreeNode.wordEnds.addElement(reader.readLine());
				reader.close();
			} catch(Exception e)
			{
				e.printStackTrace();
			}
		}
		// Initializes variable
		nodes = new Vector<TreeNode>();
		// Stores "end" if input-string is empty
		if(data.equals("")) this.data = "end";
		else
		{
			// If the input-string has some value, then the first letter is stored in this node
			this.data = String.valueOf(data.charAt(0));
			// The rest is sent on for further storage.
			insert(data.substring(1));
		}
	}
	
	/**
	 * 
	 * This method is called for inserting a new word into this tree.
	 * It will check the value in all the subsequent nodes and compare it to the first letter int the input-string.
	 * If it finds a match, it will call the insert()-method in that node and send in the input-string, minus the first letter.
	 * 
	 * @param wordPart - the received word or part-of-word
	 */
	public void insert(String wordPart)
	{
		boolean found = false;
		if(!wordPart.equals(""))
			// Finds if there is a link to a node, with a matching letter to the first letter in the input-string
			for(int i = 0; i < nodes.size(); i++)
				if(nodes.elementAt(i).getData().equals(String.valueOf(wordPart.charAt(0))))
				{
					found = true;
					// If there is a match, then that node will get a call to its insert()-method with the input-string minus the first letter.
					nodes.elementAt(i).insert(wordPart.substring(1));
					break;
				}
		// If there was no node with a match on the first letter of the input-string,
		// then this code will create a node and send in the input-string to it, and this will store the rest of the input-string.
		if(!found) nodes.addElement(new TreeNode(wordPart));
	}
	
	/**
	 * 
	 * This method recursively checks whether an input word is a word to be removed.
	 * 
	 * @param wordPart - the received word or part-of-word
	 * @return true if the word is to be removed, false if it is not to be removed.
	 */
	public boolean isStopWord(String wordPart)
	{
		// First checks if the word is not at the end
		if(!wordPart.equals(""))
		{
			// First checks if the first letter in the input equals one of the symbols registered as end-of-words-symbols
			for(int i = 0; i < TreeNode.wordEnds.size(); i++)
				if(String.valueOf(wordPart.charAt(0)).equals(wordEnds.elementAt(i)))
					// If the letter is an end-of-word-symbol, this checks if there is stored an "end" in one of the sub-nodes
					// Thus saying that this is a word to be removed and so returning the value "true".
					for(int j = 0; j < nodes.size(); j++)
						if(nodes.elementAt(j).getData().equals("end")) return true;
			// Checks if there is a node with a matching letter with the first letter in the input-string.
			// If there is a match, then the return statement is set to be the value returned from the call to
			// the "isStopWord"-method in that node where the match was found.
			// With the call to that method is sent the input-string minus the first letter.
			for(int i = 0; i < nodes.size(); i++)
				if(nodes.elementAt(i).getData().equals(String.valueOf(wordPart.charAt(0)))) return nodes.elementAt(i).isStopWord(wordPart.substring(1));
			// If the code has reached this point, it means that there is no match with the first letter of the input-string.
			// That means that the result should be "false".
			return false;
		}
		// If this code is running, it means that the inpu-string is empty.
		// That means that if one of the nodes contain "end", then the result should be "true".
		for(int i = 0; i < nodes.size(); i++)
			if(nodes.elementAt(i).getData().equals("end")) return true;
		// If this code is run it means that there has been found no "end" in any sub-node and so "false" should be returned.
		return false;
	}
	
	/**
	 * 
	 * A standard getter method
	 * 
	 * @return returns the letter of this node
	 */
	public String getData()
	{
		return data;
	}
	
}
